@@ -580,9 +580,10 @@ class MssqlCompleter(Completer):
             aliases = (tbl + str(i) for i in count(2))
         return next(a for a in aliases if normalize_ref(a) not in tbls)
 
+    # TODO: Need to account for suggestion.catalog, tbls.catalog etc
     def get_join_matches(self, suggestion, word_before_cursor):
         tbls = suggestion.table_refs
-        cols = self.populate_scoped_cols(tbls)
+        cols = self.populate_scoped_cols2(tbls)
         # Set up some data structures for efficient access
         qualified = dict((normalize_ref(t.ref), t.schema) for t in tbls)
         ref_prio = dict((normalize_ref(t.ref), n) for n, t in enumerate(tbls))
@@ -677,6 +678,7 @@ class MssqlCompleter(Completer):
 
         return self.find_matches(word_before_cursor, conds, meta='join')
 
+    # TODO: Need to account for suggestion.catalog
     def get_function_matches(
             self, suggestion, word_before_cursor, alias=False):
         if suggestion.usage == 'from':
@@ -711,10 +713,11 @@ class MssqlCompleter(Completer):
         return matches
 
     def get_schema_matches(self, suggestion, word_before_cursor):
+        conn = self.my_app.active_conn
         if suggestion.parent:
             catalog_u = self.unescape_name(suggestion.parent)
         else:
-            catalog_u = self.my_app.active_conn.current_catalog()
+            catalog_u = conn.current_catalog()
 
         catalog_e = self.escape_name(catalog_u)
         cats = self.dbmetadata['table'].keys()
@@ -724,18 +727,24 @@ class MssqlCompleter(Completer):
             schema_names_e = self.dbmetadata['table'][catalog_e].keys()
         else: 
             if suggestion.parent:
-                conn = self.my_app.active_conn
-                res = self.my_app.active_conn.find_tables(
-                        catalog = conn.sanitize_search_string(catalog_u),
-                        schema = "",
-                        table = "",
-                        type = "")
+                # Looking for schemas in a specified catalog
                 schema_names = []
-                for r in res:
-                    if (r.schema not in schema_names and r.schema != ""):
-                        schema_names.append(r.schema)
+                # Attempt list_schemas
+                schema_names = conn.list_schemas(
+                        catalog = conn.sanitize_search_string(catalog_u))
+
+                if len(schema_names) < 1:
+                    res = conn.find_tables(
+                            catalog = conn.sanitize_search_string(catalog_u),
+                            schema = "",
+                            table = "",
+                            type = "")
+                    for r in res:
+                        if (r.schema not in schema_names and r.schema != ""):
+                            schema_names.append(r.schema)
             else:
-                schema_names = set(self.my_app.active_conn.list_schemas())
+                # Looking for schemas in current catalog
+                schema_names = set(conn.list_schemas())
             
             schema_names_e = self.escaped_names(schema_names)
             if len(schema_names_e):
@@ -1086,12 +1095,35 @@ class MssqlCompleter(Completer):
 
         """
         ret = []
+        conn = self.my_app.active_conn
         if catalog is None and schema is None:
-            # fixme: need to fix this for sqlite where this is perfectly ok
+            # Query free tables, no catalog or schema (think SQLite)
+            # Trying to communicate here empty string, "" to
+            # SQLTables API.  Unfortunately, nanodbc intercepts these
+            # and translates them into nullptr which in turn are treated
+            # as wildcards per ODBC standard.  So try to work around the
+            # nanodbc intercept, and create an empty string as a
+            # null-terminated string without content.
+            res = conn.find_tables(
+                    catalog = "\x00",
+                    schema = "\x00",
+                    table = "",
+                    type = obj_type)
+
+            for r in res:
+                name_e = self.escape_name(r.name)
+                ret.append(
+                    SchemaObject(
+                        name = name_e,
+                        schema = "",
+                        catalog = ""
+                    )
+                )
+            # Don't expand dbmetadata with free tables (for now)
             return ret
         if catalog is None:
             #Set to current catalog
-            catalog = self.my_app.active_conn.current_catalog()
+            catalog = conn.current_catalog()
         # Note to self, as soon as a period is inputted a schema (parent) is
         # no longer none.  Technically we should never be in a situation where
         # catalog is not None but schema is None.
@@ -1117,7 +1149,6 @@ class MssqlCompleter(Completer):
         else:
             self.logger.debug("populate_objects(%s): Did not find %s.%s metadata.  Will query.", obj_type, catalog_e, schema_e)
             obj_names = []
-            conn = self.my_app.active_conn
             res = conn.find_tables(
                     catalog = conn.sanitize_search_string(
                         self.unescape_name(catalog)),
