@@ -3,6 +3,8 @@ import platform
 from cyanodbc import Connection
 from typing import List, Optional, Callable
 from logging import getLogger
+from asyncio import get_event_loop
+from threading import Thread
 from prompt_toolkit.layout.containers import HSplit, Window, ScrollOffsets, ConditionalContainer, Container
 from prompt_toolkit.formatted_text.base import StyleAndTextTuples
 from prompt_toolkit.formatted_text import fragment_list_width
@@ -15,13 +17,16 @@ from prompt_toolkit.layout.margins import ScrollbarMargin
 from prompt_toolkit.mouse_events import MouseEvent
 from prompt_toolkit.lexers import Lexer
 from prompt_toolkit.widgets import SearchToolbar
+from prompt_toolkit.filters import Condition
 from .conn import sqlConnection
 from .filters import ShowSidebar
 from .utils import if_mousedown
+from .__init__ import __version__
 
 class myDBObject:
     def __init__(
         self,
+        my_app: "sqlApp",
         conn: sqlConnection,
         name: str,
         otype: str,
@@ -31,6 +36,7 @@ class myDBObject:
         next_object: Optional["myDBObject"] = None
     ) -> None:
 
+        self.my_app = my_app
         self.conn = conn
         self.children = children
         self.parent = parent
@@ -41,12 +47,37 @@ class myDBObject:
         self.selected: bool = False
         self.logger = getLogger(__name__)
 
-    def expand(self) -> None:
+    def _expand_internal(self) -> None:
         """
         Populates children and sets parent for children nodes
         """
         raise NotImplementedError()
 
+    def expand(self) -> None:
+        """
+        Populates children and sets parent for children nodes
+        """
+        if self.children is not None:
+            return None
+
+        loop = get_event_loop()
+        self.my_app.show_expanding_object = True
+        self.my_app.application.invalidate()
+        def _redraw_after_io():
+            """ Callback, scheduled after threaded I/O
+                completes """
+            self.my_app.show_expanding_object = False
+            self.my_app.application.invalidate()
+
+        def _run():
+            """ Executes in a thread """
+            self._expand_internal() # Blocking I/O
+            loop.call_soon_threadsafe(_redraw_after_io)
+
+        # (Don't use 'run_in_executor', because daemon is ideal here.
+        t = Thread(target = _run, daemon = True)
+        t.start()
+
     def collapse(self) -> None:
         """
         Populates children and sets parent for children nodes
@@ -54,18 +85,16 @@ class myDBObject:
         next_object.  This way we re-query the database / force re-fresh
         which may be suboptimal.  TODO: Codify not/refresh path
         """
-        if not self.selected:
+        if self is not self.my_app.selected_object:
             return
         if self.children is not None:
-            self.next_object = self.children[len(self.children) - 1].next_object
             obj = self.children[len(self.children) - 1].next_object
-            if obj is not None and obj.level <= self.level:
-                self.next_object = obj
-            else:
-                self.next_object = None
+            while obj.level > self.level:
+                obj = obj.next_object
+            self.next_object = obj
             self.children = None
         elif self.parent is not None:
-            self.parent.selected = True
+            self.my_app.selected_object = self.parent
             self.parent.collapse()
 
     def add_children(self, list_obj: List["myDBObject"]) -> None:
@@ -75,49 +104,12 @@ class myDBObject:
         self.children[len(self.children) - 1].next_object = self.next_object
         self.next_object = self.children[0]
 
-    def select_next(self) -> None:
-        if self.selected_object is None or self.next_object is None:
-            self.selected = True
-            return None
-
-        if self.selected:
-            self.next_object.selected = True
-            self.selected = False
-        else:
-            self.next_object.select_next()
-
-    def select_previous(self) -> None:
-        if self.selected == True:
-        # Can't proceed past this point so exit
-        # should never happen unless the very first object
-        # is seleced in which case it makes sense
-            return None
-        if self.selected_object is None or self.next_object is None:
-            self.selected = True
-            return None
-
-        if self.next_object.selected == True:
-            self.next_object.selected = False
-            self.selected = True
-        else:
-            self.next_object.select_previous()
-
-    @property
-    def selected_object(self) -> "myDBObject":
-        if self.selected:
-            return self
-        elif self.next_object:
-            return self.next_object.selected_object
+class myDBColumn(myDBObject):
+    def _expand_internal(self) -> None:
         return None
 
-class myDBColumn(myDBObject):
-    def expand(self) -> None:
-        self.children = None
-
 class myDBTable(myDBObject):
-    def expand(self) -> None:
-        if self.children is not None:
-            return None
+    def _expand_internal(self) -> None:
         cat = "%"
         schema = "%"
         # https://docs.microsoft.com/en-us/sql/odbc/reference/syntax/sqlcolumns-function?view=sql-server-ver15
@@ -139,6 +131,7 @@ class myDBTable(myDBObject):
                 column = "%")
 
         lst = [myDBColumn(
+            my_app = self.my_app,
             conn = self.conn,
             name = col.column,
             otype = col.type_name,
@@ -147,10 +140,11 @@ class myDBTable(myDBObject):
         if len(lst):
             self.add_children(list_obj = lst)
 
+        return None
+
 class myDBSchema(myDBObject):
-    def expand(self) -> None:
-        if self.children is not None:
-            return None
+    def _expand_internal(self) -> None:
+
         cat = self.conn.sanitize_search_string(self.parent.name) if self.parent is not None else "%"
         res = self.conn.find_tables(
                 catalog = cat,
@@ -158,6 +152,7 @@ class myDBSchema(myDBObject):
                 table = "",
                 type = "")
         lst = [myDBTable(
+            my_app = self.my_app,
             conn = self.conn,
             name = table.name,
             otype = table.type,
@@ -166,88 +161,93 @@ class myDBSchema(myDBObject):
         if len(lst):
             self.add_children(list_obj = lst)
 
-class myDBCatalog(myDBObject):
-    def expand(self) -> None:
-        if self.children is not None:
-            return None
+        return None
 
-        res = self.conn.find_tables(
-                catalog = self.conn.sanitize_search_string(self.name),
-                schema = "",
-                table = "",
-                type = "")
-        schemas = []
-        for r in res:
-            if (r.schema not in schemas and r.schema != ""):
-                schemas.append(r.schema)
+class myDBCatalog(myDBObject):
+    def _expand_internal(self) -> None:
+        schemas = lst = []
+        schemas = self.conn.list_schemas(
+                catalog = self.conn.sanitize_search_string(self.name))
+
+        if len(schemas) < 1:
+            res = self.conn.find_tables(
+                    catalog = self.conn.sanitize_search_string(self.name),
+                    schema = "",
+                    table = "",
+                    type = "")
+            for r in res:
+                if (r.schema not in schemas and r.schema != ""):
+                    schemas.append(r.schema)
         if len(schemas):
             lst = [myDBSchema(
+                my_app = self.my_app,
                 conn = self.conn,
                 name = schema,
                 otype = "Schema",
                 parent = self,
                 level = self.level + 1) for schema in schemas]
-            self.add_children(list_obj = lst)
-            return None
-
-        # No schemas found; but if there are tables then these are direct
-        # descendents
-        lst = [myDBTable(
-            conn = self.conn,
-            name = table.name,
-            otype = table.type,
-            parent = self,
-            level = self.level + 1) for table in res]
+        else:
+            # No schemas found; but if there are tables then these are direct
+            # descendents, i.e. MySQL
+            lst = [myDBTable(
+                my_app = self.my_app,
+                conn = self.conn,
+                name = table.name,
+                otype = table.type,
+                parent = self,
+                level = self.level + 1) for table in res]
         if len(lst):
             self.add_children(list_obj = lst)
 
+        return None
+
+
 class myDBConn(myDBObject):
-    def expand(self) -> None:
+    def _expand_internal(self) -> None:
         if not self.conn.connected():
             return None
-        if self.children is not None:
-            return None
 
+        lst = []
         cat_support = self.conn.catalog_support()
         if cat_support:
             rows = self.conn.list_catalogs()
             if len(rows):
                 lst = [myDBCatalog(
+                    my_app = self.my_app,
                     conn = self.conn,
                     name = row,
                     otype = "Catalog",
                     parent = self,
                     level = self.level + 1) for row in rows]
-                self.add_children(list_obj = lst)
-            return None
-
-        res = self.conn.find_tables(
-                catalog = "%",
-                schema = "",
-                table = "",
-                type = "")
-        schemas = []
-        for r in res:
-            if (r.schema not in schemas and r.schema != ""):
-                schemas.append(r.schema)
-        if len(schemas):
-            lst = [myDBSchema(
-                conn = self.conn,
-                name = schema,
-                otype = "Schema",
-                parent = self,
-                level = self.level + 1) for schema in schemas]
-            self.add_children(list_obj = lst)
-            return None
-
-        lst = [myDBTable(
-            conn = self.conn,
-            name = table.name,
-            otype = table.type,
-            parent = self,
-            level = self.level + 1) for table in res]
+        else:
+            schemas = []
+            res = self.conn.find_tables(
+                    catalog = "%",
+                    schema = "",
+                    table = "",
+                    type = "")
+            for r in res:
+                if (r.schema not in schemas and r.schema != ""):
+                    schemas.append(r.schema)
+            if len(schemas):
+                lst = [myDBSchema(
+                    my_app = self.my_app,
+                    conn = self.conn,
+                    name = schema,
+                    otype = "Schema",
+                    parent = self,
+                    level = self.level + 1) for schema in schemas]
+            else:
+                lst = [myDBTable(
+                    my_app = self.my_app,
+                    conn = self.conn,
+                    name = table.name,
+                    otype = table.type,
+                    parent = self,
+                    level = self.level + 1) for table in res]
         if len(lst):
             self.add_children(list_obj = lst)
+        return None
 
 def sql_sidebar_help(my_app: "sqlApp"):
     """
@@ -259,7 +259,7 @@ def sql_sidebar_help(my_app: "sqlApp"):
         """
         Return the description of the selected option.
         """
-        obj = my_app.obj_list[0].selected_object
+        obj = my_app.selected_object
         if obj is not None:
             return obj.name
         return ""
@@ -271,8 +271,33 @@ def sql_sidebar_help(my_app: "sqlApp"):
         content=Window(
             FormattedTextControl(get_help_text), style=token, height=Dimension(min=3)
             ),
-        filter=ShowSidebar(my_app) & ~is_done,
-        )
+        filter = ~is_done
+        & ShowSidebar(my_app)
+        & Condition(
+            lambda: not my_app.show_exit_confirmation
+        ))
+
+def expanding_object_notification(my_app: "sqlApp"):
+    """
+    Create the `Layout` for the 'Expanding object' notification.
+    """
+
+    def get_text_fragments():
+        # Show navigation info.
+        return [("fg:red", "Expanding object ...")]
+
+    return ConditionalContainer(
+        content = Window(
+            FormattedTextControl(get_text_fragments),
+            style = "class:sidebar",
+            width=Dimension.exact( 45 ),
+            height=Dimension(max = 1),
+        ),
+        filter = ~is_done
+        & ShowSidebar(my_app)
+        & Condition(
+            lambda: my_app.show_expanding_object
+        ))
 
 def sql_sidebar_navigation():
     """
@@ -282,23 +307,23 @@ def sql_sidebar_navigation():
     def get_text_fragments():
         # Show navigation info.
         return [
-            ("class:sidebar", "   "),
-            ("class:sidebar.key", "[Up/Dn]"),
-            ("class:sidebar", " "),
-            ("class:sidebar.description", "Navigate"),
-            ("class:sidebar", " "),
-            ("class:sidebar.key", "[L/R]"),
-            ("class:sidebar", " "),
-            ("class:sidebar.description", "Expand/Collapse"),
-            ("class:sidebar", "\n   "),
-            ("class:sidebar.key", "[Enter]"),
-            ("class:sidebar", " "),
-            ("class:sidebar.description", "Connect/Preview"),
+            ("class:sidebar.navigation", "   "),
+            ("class:sidebar.navigation.key", "[Up/Dn]"),
+            ("class:sidebar.navigation", " "),
+            ("class:sidebar.navigation.description", "Navigate"),
+            ("class:sidebar.navigation", " "),
+            ("class:sidebar.navigation.key", "[L/R]"),
+            ("class:sidebar.navigation", " "),
+            ("class:sidebar.navigation.description", "Expand/Collapse"),
+            ("class:sidebar.navigation", "\n   "),
+            ("class:sidebar.navigation.key", "[Enter]"),
+            ("class:sidebar.navigation", " "),
+            ("class:sidebar.navigation.description", "Connect/Preview"),
         ]
 
     return Window(
         FormattedTextControl(get_text_fragments),
-        style = "class:sidebar",
+        style = "class:sidebar.navigation",
         width=Dimension.exact( 45 ),
         height=Dimension(max = 2),
     )
@@ -320,11 +345,7 @@ def show_sidebar_button_info(my_app: "sqlApp") -> Container:
             ("class:status-toolbar.key", "[C-t]", toggle_sidebar),
             ("class:status-toolbar", " Object Browser", toggle_sidebar),
             ("class:status-toolbar", " - "),
-            (
-                "class:status-toolbar.python-version",
-                "%s %i.%i.%i"
-                % (platform.python_implementation(), version[0], version[1], version[2]),
-                ),
+            ("class:status-toolbar.cli-version", "odbcli %s" % __version__),
             ("class:status-toolbar", " "),
             ]
     width = fragment_list_width(tokens)
@@ -341,6 +362,9 @@ def show_sidebar_button_info(my_app: "sqlApp") -> Container:
                 width=Dimension.exact(width),
                 ),
             filter=~is_done
+            & Condition(
+                lambda: not my_app.show_exit_confirmation
+            )
 #            & renderer_height_is_known
             )
 
@@ -356,8 +380,9 @@ def sql_sidebar(my_app: "sqlApp") -> Window:
     def tokenize_obj(obj: "myDBObject") -> StyleAndTextTuples:
         " Recursively build the token list "
         tokens: StyleAndTextTuples = []
-        selected = obj.selected
+        selected = obj is my_app.selected_object
         expanded = obj.children is not None
+        connected = obj.otype == "Connection" and obj.conn.connected()
         active = my_app.active_conn is not None and my_app.active_conn is obj.conn and obj.level == 0
 
         act = ",active" if active else ""
@@ -367,13 +392,12 @@ def sql_sidebar(my_app: "sqlApp") -> Window:
         else:
             name_trim = ("%-" + str(24 - 2 * obj.level) + "s") % obj.name
 
+        tokens.append(("class:sidebar.label" + sel + act, " >" if connected else "  "))
         tokens.append(
-            ("class:sidebar.status" + sel, " " * 2 * obj.level, expand_item)
+            ("class:sidebar.label" + sel, " " * 2 * obj.level, expand_item)
         )
-        tokens.append(("class:sidebar" + sel, " >" if selected else "  "))
         tokens.append(
             ("class:sidebar.label" + sel + act,
-#                ("%-" + str(24 - 2 * obj.level) + "s") % obj.name,
             name_trim,
             expand_item)
         )
@@ -384,10 +408,12 @@ def sql_sidebar(my_app: "sqlApp") -> Window:
             tokens.append(("[SetCursorPosition]", ""))
 
         if expanded:
-            tokens.append(("class:sidebar", "\/"))
+            tokens.append(("class:sidebar.status" + sel + act, "\/"))
         else:
-            tokens.append(("class:sidebar", " <" if selected else "  "))
+            tokens.append(("class:sidebar.status" + sel + act, " <" if selected else "  "))
 
+        # Expand past the edge of the visible buffer to get an even panel
+        tokens.append(("class:sidebar.status" + sel + act, " " * 10))
         return tokens
 
     def _buffer_pos_changed(buff):
@@ -401,9 +427,6 @@ def sql_sidebar(my_app: "sqlApp") -> Window:
             if line_no < 0:  # When the cursor is above the inserted region.
                 raise IndexError
 
-            obj = my_app.obj_list[0].selected_object
-            if obj is not None:
-                obj.selected = False
             idx = 0
             obj = my_app.obj_list[0]
             while idx < line_no:
@@ -412,7 +435,7 @@ def sql_sidebar(my_app: "sqlApp") -> Window:
                 idx += 1
                 obj = obj.next_object
 
-            obj.selected = True
+            my_app.selected_object = obj
 
         except IndexError:
             pass
@@ -448,12 +471,12 @@ def sql_sidebar(my_app: "sqlApp") -> Window:
     class myControl(BufferControl):
 
         def move_cursor_down(self):
-            my_app.obj_list[0].select_next()
+            my_app.select_next()
         # Need to figure out what do do here
         # AFAICT thse are only called for the mouse handler
         # when events are otherwise not handled
         def move_cursor_up(self):
-            my_app.obj_list[0].select_previous()
+            my_app.select_previous()
 
         def mouse_handler(self, mouse_event: MouseEvent) -> "NotImplementedOrNone":
             """
@@ -471,12 +494,12 @@ def sql_sidebar(my_app: "sqlApp") -> Window:
             obj = my_app.obj_list[0]
             res.append(obj.name)
             res_tokens.append(tokenize_obj(obj))
-            found_selected = obj.selected
+            found_selected = obj is my_app.selected_object
             idx = 0
-            while obj.next_object:
+            while obj.next_object is not my_app.obj_list[0]:
                 res.append(obj.next_object.name)
                 res_tokens.append(tokenize_obj(obj.next_object))
-                if not obj.selected and not found_selected:
+                if obj is not my_app.selected_object and not found_selected:
                     count += 1
                     idx += len(obj.name) + 1 # Newline character
                 else:
@@ -502,7 +525,12 @@ def sql_sidebar(my_app: "sqlApp") -> Window:
             right_margins = [ScrollbarMargin(display_arrows = True)],
             style = "class:sidebar",
             width = Dimension.exact( 45 ),
-            height = Dimension(min = 15, preferred = 33),
+            height = Dimension(min = 7, preferred = 33),
             scroll_offsets = ScrollOffsets(top = 1, bottom = 1)),
-        Window(style="class:sidebar,separator", height=1),
+        Window(
+            height = Dimension.exact(1),
+            char = "\u2500",
+            style = "class:sidebar,separator",
+            ),
+        expanding_object_notification(my_app),
         sql_sidebar_navigation()])
